/***************************************************************************
  
    Disk Based Hash library DBH

    static functions

copyright Edscott Wilson Garcia 2002-2010
published under LGPL Public license
You should have received a license copy along with this file.

**************************************************************************/
static int dbh_updateBranch (DBHashTable *, FILE_POINTER);
static unsigned char dbh_readBranches (DBHashTable *, FILE_POINTER);

static FILE_POINTER *dbh_locate (DBHashTable *, FILE_POINTER *);
static FILE_POINTER *dbh_locateFind (DBHashTable *, FILE_POINTER *, int);
static void dbh_cuenta (unsigned char *, unsigned char, FILE_POINTER);
static FILE_POINTER dbh_write (char, DBHashTable *, unsigned char);
static int dbh_read (char, DBHashTable *, unsigned char);
static int dbh_readheader (DBHashTable *);

#ifdef TURN
static FILE_POINTER dbh_turnaround (FILE_POINTER);
#endif
static FILE_POINTER dbh_z (unsigned char, unsigned char);
static unsigned char dbh_cuantumR (FILE_POINTER, unsigned char,
				   FILE_POINTER *);
static unsigned char dbh_cuantum (FILE_POINTER, unsigned char,
				  FILE_POINTER *);
static void dbh_cifra (unsigned char *, unsigned char, unsigned char,
		       FILE_POINTER *);

static int dbh_barrelong (DBHashTable *, FILE_POINTER, int);
static int dbh_reversebarrelong (DBHashTable *, FILE_POINTER, int);
static int dbh_reversebarre (DBHashTable *, FILE_POINTER, int);
static int dbh_barre (DBHashTable *, FILE_POINTER, int);

static void dbh_transfer (DBHashTable *);
static DBHashTable *dbh_open_S (const char *, int);

static int dbh_size (DBHashTable *, FILE_POINTER);
/*************************************************************************/

#if 0
// Faster without this.
static int
sync_fd (int fd) {
  /*if (fsync(fd) < 0) {
      // this should not happen, really.
      fprintf(stderr,"fsync(): %s\n", strerror(errno));
      return 0;
  }*/
 return 1;
}
#endif
    
static int
advance_fp (DBHashTable * dbh, FILE_POINTER offset)
{
  //if (!sync_fd(dbh->fd)) return 0;
  if (lseek (dbh->fd, (off_t) offset, SEEK_CUR) < 0)
    {
      fprintf (stderr, "Error: advance_fp failed to advance fp %lld\n",(long long) offset);
      return 0;
    }
  return 1;
}

static int
place_fp_at (DBHashTable * dbh, FILE_POINTER seek)
{
  //if (!sync_fd(dbh->fd)) return 0;
  if (lseek (dbh->fd, (off_t) seek, SEEK_SET) != seek)
    {
      // this should not happen...
      fprintf (stderr,
	       "Error: dbh_locate() cannot place file pointer at bof=0x%llu\n",
	       (long long unsigned) seek);
      return 0;
    }
  return 1;
}

static FILE_POINTER
place_eof (DBHashTable * dbh)
{
  //if (!sync_fd(dbh->fd)) return 0;
  FILE_POINTER eof;
  eof = (FILE_POINTER) lseek (dbh->fd, (off_t) 0L, SEEK_END);
  if (eof < sizeof (dbh_header_t))
    {
      fprintf (stderr, "Error: tell_eof() <  sizeof (dbh_header_t)\n");
      eof = -1;
    }
  if (eof < 0)
    {
      // this should not happen...
      fprintf (stderr,
	       "Error: tell_eof() cannot place file pointer at eof\n");
    }
  return eof;
}


  /* Funciones nivel 0: */
static void
dbh_operate (DBHashTable * dbh)
{
  if (dbh == NULL)
    {
      DBG ("Must define function operate(place*) before doing a sweep\n");
    }
}





static char *
dbh_randomfilename (DBHashTable * dbh, char code)
{
  time_t segundos;
  FILE_POINTER divisor;
  char *archivo;
  char *tmpdir;
  if (!dbh) return NULL;
  // determine best tmp directory
  if (dbh->head_info->user_tmpdir == 0)
    {
      struct stat dbh_st, tmp_st;
      stat (dbh->path, &dbh_st);
      stat ("/tmp", &tmp_st);
      if (dbh_st.st_dev == tmp_st.st_dev)
	{
	  dbh->tmpdir = (char *) malloc (strlen ("/tmp") + 1);
	  strcpy (dbh->tmpdir, "/tmp");
	}
      else
	{
	  dbh->tmpdir=(char *)malloc(strlen(dbh->path)+1);
	  strcpy(dbh->tmpdir,dbh->path);
	  if (strrchr(dbh->tmpdir,'/')) *strrchr(dbh->tmpdir,'/')=0;
	  else {
	      fprintf(stderr, "%s is not an absolute path!",dbh->path);
	  }
	}
    }
  tmpdir = dbh->tmpdir;
  archivo = (char *) malloc (strlen (tmpdir) + 1 + 1 + 6 + 4 + 1);

  time (&segundos);
  srand ((unsigned) segundos);
  divisor = RAND_MAX / 10000;
  if ((segundos = rand () / divisor) > 100000L)
    segundos = 50001;
  sprintf (archivo, "%s/%c%ld.tmp", tmpdir, code, (long) segundos);

  return archivo;
}  /***************************************************************************************/


static pthread_mutex_t lock_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t unlock_mutex = PTHREAD_MUTEX_INITIALIZER;
//mode=1 --> write lock
static void
lock_fd (int fd, int mode, const char *archivo)
{
  //  return;
  // get the file lock, wait if necesary:
  struct flock lock;

  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;		// to eof  

  if (mode)
    {
      lock.l_type = F_WRLCK;
    }
  else
    {
      lock.l_type = F_RDLCK;
    }

  pthread_mutex_lock (&lock_mutex);
  if (fcntl (fd, F_SETLKW, &lock) < 0)
    {
      fprintf (stderr, "fcntl(F_SETLKW)(%s): %s\n",
	       archivo, strerror (errno));
    }
  pthread_mutex_unlock (&lock_mutex);
}

static void
unlock (int fd)
{
  //return;
  // release the lock:
  struct flock lock;

  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;		// to eof  

  lock.l_type = F_UNLCK;
  pthread_mutex_lock (&unlock_mutex);
  if (fcntl (fd, F_SETLKW, &lock) < 0)
    {
      fprintf (stderr, "fcntl(F_SETLKW): %s\n", strerror (errno));
    }
  pthread_mutex_unlock (&unlock_mutex);
}

static int
dbh_lock_write (DBHashTable * dbh)
{
  if (!dbh)
    return 0;
  TRACE ("DBH: locking file for write %s (%d)\n", dbh->path, dbh->fd);
  lock_fd (dbh->fd, 1, dbh->path);
  return 1;
}

static int
dbh_lock_read (DBHashTable * dbh)
{
  if (!dbh)
    return 0;
  TRACE ("DBH: locking file for read %s (%d)\n", dbh->path, dbh->fd);
  lock_fd (dbh->fd, 0, dbh->path);
  return 1;
}

static int
dbh_unlock (DBHashTable * dbh)
{
  TRACE ("DBH: unlocking file %s (%d))\n", dbh->path, dbh->fd);
  if (!dbh)
    return 0;
  unlock (dbh->fd);
  return 1;
}

  /* Functions: ************************************************************** */
static unsigned char
dbh_readBranches (DBHashTable * dbh, FILE_POINTER seek)
{
  FILE_POINTER *rama;
  unsigned char ramas;
  rama = dbh->newbranch;
  TRACE ("dbh_readBranches\n");

  dbh_lock_read (dbh);
  if (!place_fp_at (dbh, seek))
    return ERROR_VALUE;
  if (read (dbh->fd, &ramas, 1) != 1)
    {
      TRACE ("dbh_readBranches error 3.1\n");
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  TRACE ("dbh_readBranches ramas=%lld\n", (long long) ramas);
  if (!advance_fp (dbh, 1L + sizeof (FILE_POINTER)))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }

  if (read (dbh->fd, rama, sizeof (FILE_POINTER) * ramas) != sizeof (FILE_POINTER) * ramas)
    {
      TRACE ("dbh_readBranches error 3.2\n");
      dbh_unlock (dbh);

      return ERROR_VALUE;
    }
  dbh_unlock (dbh);
  TRACE ("read:");
  int i;
  for (i = 0; i < ramas; i++)
    TRACE (" rama[%d]=0x%llu", i, (long long unsigned) (rama[i]));
  TRACE ("\n");
#ifdef TURN
  {
    int i;
    for (i = 0; i < ramas; i++)
      rama[i] = dbh_turnaround (rama[i]);
  }
#endif
  return ramas;
} /***************************************************/

static int
dbh_updateBranch (DBHashTable * dbh, FILE_POINTER seek)
{
  TRACE ("dbh_updateBranch\n");
  FILE_POINTER *rama;
  unsigned char ramas;
  rama = dbh->newbranch;
  dbh_lock_write (dbh);
  if (!place_fp_at (dbh, seek))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  if (read (dbh->fd, &ramas, 1) != 1)
    {
      TRACE ("dbh_updateBranch read error 4.1\n");
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }

  if (!advance_fp (dbh, 1L + sizeof (FILE_POINTER)))
    return ERROR_VALUE;

#ifdef TURN
  {
    int i;
    for (i = 0; i < ramas; i++)
      rama[i] = dbh_turnaround (rama[i]);
  }
#endif
  TRACE ("write:");
  int i;
  for (i = 0; i < ramas; i++)
    TRACE (" rama[%d]=0x%llu", i, (long long unsigned) (rama[i]));
  TRACE ("\n");

  if (write (dbh->fd, rama, sizeof (FILE_POINTER) * ramas) != sizeof (FILE_POINTER) * ramas)
    {
      TRACE ("dbh_updateBranch write error 4.2\n");
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  dbh_unlock (dbh);
#ifdef TURN
  {
    int i;
    for (i = 0; i < ramas; i++)
      rama[i] = dbh_turnaround (rama[i]);
  }
#endif
  return 1;
}  /*****************************************************/

#define BRANCHOFF (dbh->head_info->n_limit-dbh->newbranches)
#define CURRENTSEEK fp[0]
#define LASTSEEK fp[1]
#define CURR_BRANCH fp[2]
#define SWITCHED (dbh->head_info->reservedD)
//FILE_POINTER *fp must have size = 3
static FILE_POINTER *
dbh_locate (DBHashTable *dbh, FILE_POINTER * fp)
{
  TRACE ("dbh_locate\n");
  FILE_POINTER currentseek, lastseek;
  int i;			/*,offset; */
  lastseek = currentseek = 0;
  CURRENTSEEK = LASTSEEK = CURR_BRANCH = 0;	/* ERROR value assigned */
  currentseek = dbh->head_info->bof;

  dbh_lock_read (dbh);
  if (!place_fp_at (dbh, currentseek))
    {
      dbh_unlock (dbh);
      return NULL;
    }

  if (!dbh_read (NEW, dbh, 1))
    {
      dbh_unlock (dbh);
      return fp;
    }
loop:
#define OFFSET (i+(dbh->head_info->n_limit-dbh->newbranches))
  for (i = 0; i < dbh->newbranches; i++)
    {
/*    offset=i+(dbh->head_info->n_limit-dbh->newbranches);
    if (*(dbh->key+offset) != *(dbh->newkey+offset))*/
      if (*(dbh->key + OFFSET) != *(dbh->newkey + OFFSET))
/*    if ((dbh->key)[OFFSET] ^ (dbh->newkey)[OFFSET])   */
	{
	  lastseek = currentseek;
	  if ((currentseek = *(dbh->newbranch + i)) == 0)
	    {
	      CURR_BRANCH = i;
	      break;
	    }
	  if (!place_fp_at (dbh, currentseek))
	    {
	      dbh_unlock (dbh);
	      return NULL;
	    }
	  if (!dbh_read (NEW, dbh, 1))
	    {
	      dbh_unlock (dbh);
	      return fp;
	    }
	  goto loop;
	}
    }
  CURRENTSEEK = currentseek;
  LASTSEEK = lastseek;
  dbh_unlock (dbh);
  return fp;
}

static FILE_POINTER *
dbh_locateFind (DBHashTable * dbh, FILE_POINTER * fp, int keys)
{
  TRACE ("dbh_locateFind\n");
  FILE_POINTER currentseek, lastseek;
  int i;
  lastseek = currentseek = 0;
  CURRENTSEEK = LASTSEEK = CURR_BRANCH = 0;	/* ERROR value assigned */
  currentseek = dbh->head_info->bof;

  dbh_lock_read (dbh);
  if (!place_fp_at (dbh, currentseek))
    {
      dbh_unlock (dbh);
      return NULL;
    }

  if (!dbh_read (NEW, dbh, 1))
    {
      dbh_unlock (dbh);
      return fp;
    }
  if (keys > dbh->head_info->n_limit)
    keys = dbh->head_info->n_limit;
loop:
  for (i = 0; i < dbh->newbranches - (dbh->head_info->n_limit - keys); i++)
    {
      /*  offset=i+(dbh->head_info->n_limit-dbh->newbranches); */
      if (*(dbh->key + OFFSET) != *(dbh->newkey + OFFSET))
	{
	  lastseek = currentseek;
	  if ((currentseek = *(dbh->newbranch + i)) == 0)
	    {
	      CURR_BRANCH = i;
	      break;
	    }
	  if (!place_fp_at (dbh, currentseek))
	    {
	      dbh_unlock (dbh);
	      return NULL;
	    }
	  if (!dbh_read (NEW, dbh, 1))
	    {
	      dbh_unlock (dbh);
	      return fp;
	    }
	  goto loop;
	}
    }
  CURRENTSEEK = currentseek;
  LASTSEEK = lastseek;
  dbh_unlock (dbh);
  return fp;
}

static DBHashTable *
dbh_open_S (const char *archivo, int mode)
{
  TRACE ("dbh_open_S\n");
  DBHashTable *dbh;
  int fd;
  if (mode)
    {
      fd = open (archivo, O_RDWR);
    }
  else {
    fd = open (archivo, O_RDONLY);
  }

  if (fd < 0)
    {
      fprintf (stderr, "dbh_open(%s): %s\n", archivo, strerror (errno));
      return NULL;
    }
  dbh = (DBHashTable *) malloc (sizeof (DBHashTable));
  if (dbh == NULL)
    {
      //      EFAULT          Bad address (POSIX.1)
      fprintf (stderr, "dbh_open_S- malloc(%lu): %s\n",
	       (long unsigned) sizeof (DBHashTable), strerror (EFAULT));
      close (fd);
      return NULL;
    }
  memset (dbh, 0, sizeof (DBHashTable));
  /////////////////////////////////////////////

  dbh->head_info = (dbh_header_t *) malloc (sizeof (dbh_header_t));
  dbh->fd = fd;
  dbh_lock_read (dbh);
  FILE_POINTER eof = place_eof (dbh);
  dbh_unlock (dbh);
  if (eof < 0)
    {
      close (fd);
      free (dbh->head_info);
      free (dbh);
      return NULL;
    }

  dbh->path = (char *) malloc (strlen (archivo) + 1);
  strcpy (dbh->path, archivo);

  if (!dbh_readheader (dbh))
    {
      close (fd);
      free (dbh->path);
      free (dbh->head_info);
      free (dbh);
      return NULL;
    }
  dbh->operate = dbh_operate;
  dbh->branch =
    (FILE_POINTER *) malloc (dbh->head_info->n_limit *
			     sizeof (FILE_POINTER));
  dbh->newbranch =
    (FILE_POINTER *) malloc (dbh->head_info->n_limit *
			     sizeof (FILE_POINTER));
  dbh->key = (unsigned char *) malloc (dbh->head_info->n_limit);
  dbh->newkey = (unsigned char *) malloc (dbh->head_info->n_limit);
  if (mode == WRITE)
    dbh->head_info->writeOK = 1;
  else
    dbh->head_info->writeOK = 0;
  //dbh_lock_write(dbh);
  dbh_size (dbh, (int) dbh->head_info->record_length);
  TRACE ("DBH: opened %s \n", archivo);
  return dbh;
}  /****************************************************************************/

static void
dbh_cuenta (unsigned char *numero, unsigned char orden, FILE_POINTER cuanto)
{
  unsigned char q;
  FILE_POINTER m;
  if (orden == 0)
    numero[0] = 0;
  if (orden == 1)
    numero[0] = (unsigned char) cuanto;
  q = dbh_cuantumR (cuanto, orden, &m);	/* en que cuantum estamos */
  orden--;
  dbh_cifra (numero, q, orden, &m);
/* dbh_cifra(numero,q,orden-1,&m);*/
} /***********************************************************************/

#define RECORD_EXTENT  ( 2 + sizeof(FILE_POINTER)*(*how_many_branches+1) + dbh->head_info->n_limit + the_user_databytes )
static FILE_POINTER
dbh_write (char newinfo, DBHashTable * dbh, unsigned char write_branches)
{
  TRACE ("dbh_write\n");
  int i;
  void *the_data;
  unsigned char *how_many_branches;
  FILE_POINTER the_user_databytes;
  FILE_POINTER *the_branches;
  unsigned char *the_key;
  if (newinfo)
    {
      the_branches = dbh->newbranch;
    }
  else
    {
      the_branches = dbh->branch;
    }
  the_data = dbh->data;
  how_many_branches = &(dbh->newbranches);
  the_user_databytes = dbh->bytes_userdata;
  the_key = dbh->key;


/* only first write is condition tested. The remainder are assumed if one is passed */

  if (the_user_databytes > dbh->head_info->record_length)
    {
      fprintf (stderr,
	       "dbh->bytes_userdata (%lld) is greater than dbh->head_info->record_length (%lld). This is wrong and I stubbornly refuse to write\n",
	       (long long) dbh->bytes_userdata,
	       (long long) dbh->head_info->record_length);
      return ERROR_VALUE;
    }
  if (write (dbh->fd, how_many_branches, 1) != 1)
    return ERROR_VALUE;
  if (write (dbh->fd, &(dbh->flag), 1) != 1)
    return ERROR_VALUE;
  i = write (dbh->fd, &the_user_databytes, sizeof (FILE_POINTER));
  if (i != sizeof (FILE_POINTER))
    return ERROR_VALUE;
  if (write_branches)
    {
#ifdef TURN
      for (i = 0; i < *how_many_branches; i++)
	the_branches[i] = dbh_turnaround (the_branches[i]);
#endif
      i =
	write (dbh->fd, the_branches, sizeof (FILE_POINTER) * (*how_many_branches));
      if (i != sizeof (FILE_POINTER) * (*how_many_branches))
	return ERROR_VALUE;
#ifdef TURN
      for (i = 0; i < *how_many_branches; i++)
	the_branches[i] = dbh_turnaround (the_branches[i]);
#endif
    }
  else
    {
      if (!advance_fp (dbh, *how_many_branches * sizeof (FILE_POINTER)))
	return ERROR_VALUE;

    }
  if (write (dbh->fd, the_key, dbh->head_info->n_limit) != dbh->head_info->n_limit)
    {
      fprintf (stderr, "fwrite: %s\n", strerror (errno));
      return ERROR_VALUE;
    }
  if (the_user_databytes)
    {
      if (dbh->head_info->reservedC)
	the_user_databytes =
	  write (dbh->fd, the_data, dbh->head_info->record_length);
      else
	the_user_databytes =
	  write (dbh->fd, the_data, the_user_databytes);
      return (the_user_databytes);
    }
  else
    return (ERROR_VALUE);
}  /************************************************************************************/

static int
dbh_read (char newinfo, DBHashTable * dbh, unsigned char read_data)
{
  TRACE ("dbh_read\n");
  unsigned char keylength;
  void *the_data;
  unsigned char *how_many_branches;
  FILE_POINTER *the_user_databytes;
  FILE_POINTER *the_branches;
  unsigned char *the_key;
  if (newinfo)
    {
      the_branches = dbh->newbranch;
      the_data = dbh->newdata;
      how_many_branches = &(dbh->newbranches);
      the_user_databytes = &(dbh->newbytes_userdata);
      the_key = dbh->newkey;
    }
  else
    {
      the_branches = dbh->branch;
      the_data = dbh->data;
      how_many_branches = &(dbh->branches);
      the_user_databytes = &(dbh->bytes_userdata);
      the_key = dbh->key;
    }
  keylength = dbh->head_info->n_limit;


  if (read (dbh->fd, how_many_branches, 1) != 1)
    {
      DBG ("fread error 2.1 (on empty dbh file this is OK)\n");
      return ERROR_VALUE;
    }
  if (read (dbh->fd, &(dbh->flag), 1) != 1)
    {

      DBG ("fread error 2.2\n");
      return ERROR_VALUE;
    }
  if (read (dbh->fd, the_user_databytes, sizeof (FILE_POINTER)) != sizeof (FILE_POINTER))
    {
      DBG ("fread error 2.3\n");

      return ERROR_VALUE;
    }

  if (*the_user_databytes > dbh->head_info->record_length)
    {
      fprintf (stderr,
	       "dbh_read(): the_user_databytes (%ld) is greater than dbh->head_info->record_length (%lld). This is wrong and I stubbornly refuse to read\n",
	       (long) the_user_databytes,
	       (long long) (dbh->head_info->record_length));
      return ERROR_VALUE;
    }
  if (*the_user_databytes == 0){
      fprintf (stderr,
	       "dbh_RECORD_SIZE() == 0. This is absurd. Use dbh_set_recordsize ()\n");
      return ERROR_VALUE;
  }
  if (read
      (dbh->fd, the_branches, sizeof (FILE_POINTER) * (*how_many_branches)) != sizeof (FILE_POINTER) * (*how_many_branches))
    {
      DBG ("fread error 2.4\n");

      return ERROR_VALUE;
    }
#ifdef TURN
  {
    int i;
    for (i = 0; i < *how_many_branches; i++)
      the_branches[i] = dbh_turnaround (the_branches[i]);
  }
#endif
  if (read (dbh->fd, the_key, keylength) != keylength)
    {
      DBG ("fread error 2.5\n");
      return ERROR_VALUE;
    }
  if (read_data)
    {
      if (read (dbh->fd, the_data, *the_user_databytes) != *the_user_databytes)
	{
	  DBG ("fread error 2.51: blocksize=%lld\n",(long long)(*the_user_databytes));

	  return ERROR_VALUE;
	}
    }
  return 1;
} /************************************************************************************/

static int
dbh_readheader (DBHashTable * dbh)
{
  TRACE ("dbh_readheader\n");
  dbh_lock_read (dbh);
  if (!place_fp_at (dbh, 0L))
    {
      dbh_unlock (dbh);
      return 0;
    }
  if (read (dbh->fd, dbh->head_info, sizeof (dbh_header_t)) != sizeof (dbh_header_t))
    {
      fprintf (stderr,
	       "Failed to read header for %s: %s\n",dbh->path, strerror(errno));
      dbh_unlock (dbh);
      return 0;
    }
  dbh_unlock (dbh);
  if (strncmp (dbh->head_info->version, DBH_FILE_VERSION, 15) != 0)
    {
      fprintf (stderr,
	       "Failed to read header for %s at dbh_readheader(): strncmp (\"%s\",\"%s\")\n",
	       dbh->path, dbh->head_info->version, DBH_FILE_VERSION);
      return 0;
    }
  /* volatile values: */
  dbh->head_info->user_tmpdir=0;

#ifdef TURN
  dbh->head_info->bof = dbh_turnaround (dbh->head_info->bof);
  dbh->head_info->record_length =
    dbh_turnaround (dbh->head_info->record_length);
  dbh->head_info->total_space =
    dbh_turnaround (dbh->head_info->total_space);
  dbh->head_info->data_space = dbh_turnaround (dbh->head_info->data_space);
  dbh->head_info->erased_space =
    dbh_turnaround (dbh->head_info->erased_space);
  dbh->head_info->records = dbh_turnaround (dbh->head_info->records);
#endif
  return 1;
}  /************************************************************************************/

#ifdef TURN
static FILE_POINTER
dbh_turnaround (FILE_POINTER x)
{
  char *where, temp;
  where = (char *) &x;
  temp = where[0];
  where[0] = where[3];
  where[3] = temp;
  temp = where[1];
  where[1] = where[2];
  where[2] = temp;
  return x;
} /************************************************************************************/
#endif

static FILE_POINTER
dbh_z (unsigned char n, unsigned char m)	/* n=cuantum m=orden */
{
  FILE_POINTER s = 0;
  unsigned char i;
  if (m < 2)
    return 1;
  if (m == 2)
    return (n + 1);
  m--;
  for (i = n; i > 0; i--)
    s += dbh_z (i, m);
/* for (i=n;i>0;i--) s += dbh_z(i,m-1);*/
  return (s + 1);
}  /************************************************************************************/

static unsigned char
dbh_cuantumR (FILE_POINTER cuanto, unsigned char orden, FILE_POINTER * m)
{
  unsigned char i;
  FILE_POINTER ztotal, grandtotal = 0;
  for (i = 0; 1; i++)
    {
      if (i == orden)
	ztotal = grandtotal;
      else
	ztotal = dbh_z (i, orden);
      grandtotal += ztotal;
      if (cuanto <= grandtotal)
	{
	  if (m)
	    *m = grandtotal - cuanto + 1;
	  break;
	}
    }
  return i;
}  /************************************************************************************/

static unsigned char
dbh_cuantum (FILE_POINTER cuanto, unsigned char orden, FILE_POINTER * m)
{
  unsigned char i;
  FILE_POINTER ztotal, grandtotal = 0;
  for (i = 0; 1; i++)
    {
      if (i == orden)
	ztotal = grandtotal;
      else
	ztotal = dbh_z (i, orden);
      grandtotal += ztotal;
      if (cuanto <= grandtotal)
	break;
      if (m)
	*m = cuanto - grandtotal;
    }
  return i;
}  /************************************************************************************/

static void
dbh_cifra (unsigned char *numero, unsigned char q, unsigned char orden,
	   FILE_POINTER * m)
{
  unsigned char subcuantum;
  int t;
  if (orden == 1)
    {
      t = q - (*m - 1);
      numero[0] = (unsigned char) t;
/*  numero[0]=q - (*m-1);*/
      t = *m - 1;
      numero[1] = (unsigned char) t;
/*  numero[1]= *m-1;*/
      return;
    }
  subcuantum = dbh_cuantum (*m, orden, m);
  t = q - subcuantum;
  numero[0] = (unsigned char) t;
/* numero[0]=q-subcuantum;*/
  orden--;
  dbh_cifra (numero + 1, subcuantum, orden, m);
  /* dbh_cifra(numero+1,subcuantum,orden-1,m); */
  return;
}  /************************************************************************************/


static int
dbh_barrelong (DBHashTable * dbh, FILE_POINTER startadd, int ramas)
{
  unsigned char i;
  if (dbh_load_address (dbh, startadd) == ERROR_VALUE)
    return ERROR_VALUE;
  if (dbh->head_info->dbh_exit)
    return 2;
  if (!ERASED || (ERASED && dbh->head_info->sweep_erased))
    (*(dbh->operate)) (dbh);
  if (ramas)
    {
      if ((ramas = dbh->branches - ramas) < 0)
	ramas = 0;
    }
  for (i = dbh->branches; i > ramas; i--)
    {
      if (dbh->branch[i - 1])
	dbh_barrelong (dbh, dbh->branch[i - 1], 0);
      dbh_load_address (dbh, startadd);
    }
  return 1;
}    /**********************************************************************/

static int
dbh_reversebarrelong (DBHashTable * dbh, FILE_POINTER startadd, int ramas)
{
  int i;
  if (dbh_load_address (dbh, startadd) == ERROR_VALUE)
    return ERROR_VALUE;
  if (dbh->head_info->dbh_exit)
    return 2;
  if (!ERASED || (ERASED && dbh->head_info->sweep_erased))
    (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
  if (ramas)
    {
      if ((ramas = dbh->branches - ramas) < 0)
	ramas = 0;
    }
  for (i = ramas; i < dbh->branches; i++)
    {
      if (dbh->branch[i])
	dbh_reversebarrelong (dbh, dbh->branch[i], 0);
      dbh_load_address (dbh, startadd);
    }
  return 1;
}   /******************************************************************************/

static int
dbh_reversebarre (DBHashTable * dbh, FILE_POINTER startadd, int ramas)
{
  int i;
  unsigned char oldbranches;
  FILE_POINTER *oldbranch;
  char fullmemory = 0;
  if (dbh_load_address (dbh, startadd) == ERROR_VALUE)
    return ERROR_VALUE;
  if (dbh->head_info->dbh_exit)
    return 2;
  if (!ERASED || (ERASED && dbh->head_info->sweep_erased))
    (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
  oldbranch =
    (FILE_POINTER *) malloc (dbh->branches * sizeof (FILE_POINTER));
  if (oldbranch == NULL)
    fullmemory = 1;
  oldbranches = dbh->branches;
  if (ramas)
    {
      if ((ramas = dbh->branches - ramas) < 0)
	ramas = 0;
    }
  if (!fullmemory)
    {
      for (i = 0; i < oldbranches; i++)
	oldbranch[i] = dbh->branch[i];
      for (i = ramas; i < oldbranches; i++)
	if (oldbranch[i])
	  dbh_reversebarre (dbh, oldbranch[i], 0);
      free (oldbranch);
    }
  else
    for (i = ramas; i < oldbranches; i++)
      {
	if (dbh->branch[i])
	  dbh_reversebarrelong (dbh, dbh->branch[i], 0);
	dbh_load_address (dbh, startadd);
      }
  return 1;
}  /******************************************************************************/

static int
dbh_barre (DBHashTable * dbh, FILE_POINTER startadd, int ramas)
{
  unsigned char i;
  FILE_POINTER *oldbranch;
  char fullmemory = 0;
  if (dbh_load_address (dbh, startadd) == ERROR_VALUE)
    return ERROR_VALUE;
  if (dbh->head_info->dbh_exit)
    return 2;
#if 0
  printf ("barre: currentseek at %lu\n", (long unsigned) startadd);
  printf ("record is erased? -> %d\n", ERASED);
#endif
  if (!ERASED || (ERASED && dbh->head_info->sweep_erased))
    (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
  oldbranch =
    (FILE_POINTER *) malloc (dbh->branches * sizeof (FILE_POINTER));
  if (oldbranch == NULL)
    fullmemory = 1;
  if (ramas)
    {
      if ((ramas = dbh->branches - ramas) < 0)
	ramas = 0;
    }
  if (!fullmemory)
    {
      for (i = 0; i < dbh->branches; i++)
	oldbranch[i] = dbh->branch[i];
      for (i = dbh->branches; i > ramas; i--)
	if (oldbranch[i - 1])
	  dbh_barre (dbh, oldbranch[i - 1], 0);
      free (oldbranch);
    }
  else
    for (i = dbh->branches; i > ramas; i--)
      {
	if (dbh->branch[i - 1])
	  dbh_barrelong (dbh, dbh->branch[i - 1], 0);
	dbh_load_address (dbh, startadd);
      }
  return 1;
}   /************************************************************************/


  /* Functions: */
static void
dbh_transfer (DBHashTable *srcdbh)
{
  DBHashTable *dbh_desdbh = srcdbh->dbh_desdbh;
  dbh_desdbh->bytes_userdata = srcdbh->bytes_userdata;
  dbh_update (dbh_desdbh);
/* if ((count++ % 100)==0) printf("\n %ld registros procesados",count);*/
  return;
} /***************************************************************************/


/*****************************************************************************/
/***    old   non static functions ***/
/*****************************************************************************/

static int
dbh_size (DBHashTable * dbh, FILE_POINTER record_length)
{

  /* despues de abrir DBHashTable */
  if (dbh == NULL)
    {
      fprintf (stderr, "dbh_size(): dbh != NULL not met\n");
      return 0;
    }
  if (dbh->head_info == NULL)
    {
      fprintf (stderr, "dbh_size(): dbh->head_info != NULL not met\n");
      return 0;
    }

  if (record_length < dbh->head_info->record_length)
    {
      TRACE ("dbh_size(): may not shrink! current size=%lld requested size=%d\n",
	       (long long) dbh->head_info->record_length, record_length);
      return 0;
    }


  void *data = malloc (record_length);
  void *newdata = malloc (record_length);
  if ((data == NULL) || (newdata == NULL))
    {
      fprintf (stderr, "dbh_size: malloc failed!\n");
      return ERROR_VALUE;
    }
  memset (data, 0, record_length);
  memset (newdata, 0, record_length);
  if (dbh->data != NULL)
    {
      memcpy (data, dbh->data, dbh->head_info->record_length);
      free (dbh->data);
    }
  if (dbh->newdata != NULL)
    {
      memcpy (newdata, dbh->newdata, dbh->head_info->record_length);
      free (dbh->newdata);
    }
  dbh->data = data;
  dbh->newdata = newdata;


  dbh->head_info->record_length = record_length;

  if (dbh->head_info->writeOK)
    dbh_writeheader (dbh);
  return 1;
} /****************************************************************************************/

static int
dbh_newbarre (DBHashTable * dbh, unsigned char *key1, unsigned char *key2,
	      unsigned char keylength)
{
  int i;
  FILE_POINTER pointer;
  if (!dbh)
    {
      fprintf (stderr, "dbh_newbarre(): %s\n", strerror (EBADF));

      return ERROR_VALUE;
    }

  dbh->head_info->dbh_exit = 0;
  if (key1 == NULL)
    {
      dbh_barre (dbh, dbh->head_info->bof, 0);
      return 1;
    }
  if (key2 != NULL)
    {
      memcpy ((void *) (dbh->key), (void *) key1, dbh->head_info->n_limit);

      for (i = 0; i < dbh->head_info->n_limit; i++)
	if (key1[i] != key2[i])
	  break;
      /* que pedo con ERASED?? (load returns errorvalue on ERASED */
      pointer = dbh_load (dbh);
      if ((pointer == ERROR_VALUE) && !(ERASED))
	return ERROR_VALUE;
/*  if ((pointer=dbh_load(dbh)) == ERROR_VALUE) return ERROR_VALUE;*/
      /* aqui hay que considerar que el nodo puede estar marcado como ERASED */
      else
	{
	  pointer = dbh_find (dbh, i);
	  dbh_barre (dbh, pointer, dbh->head_info->n_limit - i);
	}
    }
  else if (keylength)
    {
      memcpy ((void *) (dbh->key), (void *) key1, keylength);
      pointer = dbh_find (dbh, keylength);
      if (pointer == ERROR_VALUE)
	return ERROR_VALUE;
      dbh_barre (dbh, pointer, dbh->head_info->n_limit - keylength);
    }
  return 1;
} /***************************************************************************************/

static int
dbh_newreversebarre (DBHashTable * dbh, unsigned char *key1,
		     unsigned char *key2, unsigned char keylength)
{
  int i;
  FILE_POINTER pointer;
  if (!dbh)
    {
      fprintf (stderr, "dbh_newreversebarre(): %s\n", strerror (EBADF));

      return ERROR_VALUE;
    }

  dbh->head_info->dbh_exit = 0;
  if (key1 == NULL)
    {
      dbh_reversebarre (dbh, dbh->head_info->bof, 0);
      return 1;
    }
  if (key2 != NULL)
    {
      memcpy ((void *) (dbh->key), (void *) key1, dbh->head_info->n_limit);
      for (i = 0; i < dbh->head_info->n_limit; i++)
	if (key1[i] != key2[i])
	  break;
      /* que pedo con ERASED?? (load returns errorvalue on ERASED */
      pointer = dbh_load (dbh);
      if ((pointer == ERROR_VALUE) && !(ERASED))
	return ERROR_VALUE;
/*  if ((pointer=dbh_load(dbh)) == ERROR_VALUE) return ERROR_VALUE;*/
      else
	{
	  pointer = dbh_find (dbh, i);
	  dbh_reversebarre (dbh, pointer, i);
	}
    }
  else if (keylength != 0)
    {
      memcpy ((void *) dbh->key, (void *) key1, keylength);

      pointer = dbh_find (dbh, keylength);
      if (pointer == ERROR_VALUE)
	return ERROR_VALUE;
      dbh_reversebarre (dbh, pointer, keylength);
    }
  return 1;
}  /************************************************************************************/

static DBHashTable *
dbh_regen (DBHashTable * dbh, int sweep)
{
  void *temp, *newtemp;
  char *archivo2, *archivo;
  DBHashTable *newdbh;
  unsigned char *tempkey, *newtempkey;
  void (*fun) (DBHashTable *);

  if (!dbh)
    {
      fprintf (stderr, "dbh_regen(): %s\n", strerror (EBADF));

      return NULL;
    }
  archivo2 = dbh_randomfilename (dbh, 's');
  DBHashTable *dbh_desdbh = dbh_create (archivo2, dbh->head_info->n_limit);
  if (!dbh_desdbh){
	fprintf(stderr, "Cannot open for write temporary file: %s\n",archivo2);
	return dbh;
  }
  archivo = (char *) malloc (strlen (dbh->path) + 1);
  strcpy (archivo, dbh->path);
  dbh->dbh_desdbh = dbh_desdbh;

  tempkey = dbh_desdbh->key;
  newtempkey = dbh_desdbh->newkey;
  temp = dbh_desdbh->data;
  newtemp = dbh_desdbh->newdata;
  dbh_desdbh->key = dbh->key;
  dbh_desdbh->newkey = dbh->newkey;
  dbh_desdbh->data = dbh->data;
  dbh_desdbh->newdata = dbh->newdata;

  dbh_desdbh->head_info->record_length = dbh->head_info->record_length;
  fun = dbh->operate;

  dbh->operate = dbh_transfer;
  if (sweep) {
      dbh_newbarre (dbh, NULL, NULL, 0);
  } else {
      dbh_newreversebarre (dbh, NULL, NULL, 0);
  }

  dbh_desdbh->data = temp;
  dbh_desdbh->newdata = newtemp;
  dbh_desdbh->key = tempkey;
  dbh_desdbh->newkey = newtempkey;
  dbh_close (dbh_desdbh);
  dbh_close (dbh);

  struct stat st_old, st_new;
  stat (archivo, &st_old);
  stat (archivo2, &st_new);
#ifdef CREATE_REGEN_BACKUP
  char *archivobak = (char *) malloc (strlen (archivo) + strlen (".bak") + 1);
  sprintf (archivobak, "%s.bak", archivo);
  if (remove (archivobak) < 0)
    {
    }
	DBG("DBH: rename %s %s\n",archivo,archivobak);
  if (rename (archivo, archivobak) < 0);
  free (archivobak);
#endif
  if (st_old.st_dev == st_new.st_dev)
    {
	DBG("DBH: rename %s %s\n",archivo2,archivo);
      if (rename (archivo2, archivo) < 0)
	{
	  fprintf (stderr, "DBH: cannot write to %s: %s\n", archivo,
		   strerror (errno));
	}
    }
  else
    {
	DBG("DBH: mv %s %s\n",archivo2,archivo);
      pid_t pid = fork ();
      if (!pid)
	{
	  execlp ("mv", "mv", archivo2, archivo, NULL);
	  _exit (123);
	}
      else
	{
	  int status;
	  waitpid (pid, &status, 0);
	}
    }

  newdbh = dbh_open (archivo);
  free (archivo);
  free (archivo2);
  newdbh->operate = fun;
  return newdbh;
} 



static void
mark_erased (DBHashTable * dbh)
{
  SET_ERASED;
  /* set file pointer at record flag byte */
  dbh_lock_write (dbh);
  if (place_fp_at (dbh, (dbh->reservedB) + 1) < 0)
    {
      dbh_unlock (dbh);
      return;
    }
  /* write the flag to the file */
  /* printf("flag for erase=0x%x\n",dbh->flag); */
  if (write (dbh->fd, &(dbh->flag), 1) != 1)
    {
      fprintf (stderr,"unable to erase: %s\n", strerror(errno));
      dbh_unlock (dbh);
      return;
    }
  /*printf("erasing %s flag at %lu\n",(char *)dbh->data,
     (long unsigned)(dbh->reservedB+1)); */
  /* update file header information */
  dbh->head_info->data_space -= dbh->bytes_userdata;
  dbh->head_info->erased_space += (dbh->bytes_userdata);
  dbh_unlock (dbh);
  dbh_writeheader (dbh);
  return;
}
static void
mark_unerased (DBHashTable * dbh)
{
  SET_UNERASED;
  /* set file pointer at record flag byte */
  dbh_lock_write (dbh);
  if (place_fp_at (dbh, (dbh->reservedB) + 1) < 0)
    return;
  /* write the flag to the file */
  if (write (dbh->fd, &(dbh->flag), 1) != 1)
    {
      fprintf (stderr,"unable to unerase: %s\n", strerror(errno));
      dbh_unlock (dbh);
      return;
    }
  dbh_unlock (dbh);
  /* update file header information */
  dbh->head_info->data_space += dbh->bytes_userdata;
  dbh->head_info->erased_space -= (dbh->bytes_userdata);
  dbh_writeheader (dbh);
  return;
}

